<?php  // $Id: questiontype.php,v 1.1.2.1 2010/07/16 13:52:43 aaboyd Exp $
/**
 * The questiontype class for the multiple choice question type.
 *
 * Note, This class contains some special features in order to make the
 * question type embeddable within a multianswer (cloze) question
 *
 * @package questionbank
 * @subpackage questiontypes
 */
class question_multichoiceset_qtype extends question_multichoice_qtype {

    function name() {
        return 'multichoiceset';
    }

    function get_question_options(&$question) {
        // Get additional information from database
        // and attach it to the question object
        if (!$question->options = get_record('question_multichoiceset', 'question',
         $question->id)) {
            notify('Error: Missing question options for multichoiceset question'.$question->id.'!');
            return false;
        }

        if (!$question->options->answers = get_records_select('question_answers', 'id IN ('.$question->options->answers.')', 'id')) {
           notify('Error: Missing question answers for multichoiceset question'.$question->id.'!');
           return false;
        }

        return true;
    }

    function save_question_options($question) {
        $result = new stdClass;
        if (!$oldanswers = get_records("question_answers", "question",
                                       $question->id, "id ASC")) {
            $oldanswers = array();
        }

        // following hack to check at least two answers exist
        $answercount = 0;
        foreach ($question->answer as $key=>$dataanswer) {
            if ($dataanswer != "") {
                $answercount++;
            }
        }
        $answercount += count($oldanswers);
        if ($answercount < 2) { // check there are at least 2 answers
            $result->notice = get_string("notenoughanswers", "qtype_multichoice", "2");
            return $result;
        }

        // Insert all the new answers
        $answers = array();

        foreach ($question->answer as $key => $dataanswer) {
            if ($dataanswer != "") {
                if ($answer = array_shift($oldanswers)) {  // Existing answer, so reuse it
                    $answer->answer     = $dataanswer;
                    $answer->fraction   = $question->fraction[$key];
                    $answer->feedback   = $question->feedback[$key];
                    if (!update_record("question_answers", $answer)) {
                        $result->error = "Could not update quiz answer! (id=$answer->id)";
                        return $result;
                    }
                } else {
                    unset($answer);
                    $answer->answer   = $dataanswer;
                    $answer->question = $question->id;
                    $answer->fraction = $question->fraction[$key];
                    $answer->feedback = $question->feedback[$key];
                    if (!$answer->id = insert_record("question_answers", $answer)) {
                        $result->error = "Could not insert quiz answer! ";
                        return $result;
                    }
                }
                $answers[] = $answer->id;
            }
        }

        $update = true;
        $options = get_record("question_multichoiceset", "question", $question->id);
        if (!$options) {
            $update = false;
            $options = new stdClass;
            $options->question = $question->id;
        }

        $options->answers = implode(",",$answers);
        if(isset($question->layout)){
             $options->layout = $question->layout;
        }
        $options->answernumbering = $question->answernumbering;
        $options->shuffleanswers = $question->shuffleanswers;
        $options->correctfeedback = trim($question->correctfeedback);
        $options->incorrectfeedback = trim($question->incorrectfeedback);
        if ($update) {
            if (!update_record("question_multichoiceset", $options)) {
                $result->error = "Could not update quiz multichoiceset options! (id=$options->id)";
                return $result;
            }
        } else {
            if (!insert_record("question_multichoiceset", $options)) {
                $result->error = "Could not insert quiz multichoiceset options!";
                return $result;
            }
        }

        // delete old answer records
        if (!empty($oldanswers)) {
            foreach($oldanswers as $oa) {
                delete_records('question_answers', 'id', $oa->id);
            }
        }

        return true;
    }

    /**
    * Deletes question from the question-type specific tables
    *
    * @return boolean Success/Failure
    * @param object $question  The question being deleted
    */
    function delete_question($questionid) {
        delete_records("question_multichoiceset", "question", $questionid);
        return true;
    }

    function create_session_and_responses(&$question, &$state, $cmoptions, $attempt) {
        // create an array of answerids ??? why so complicated ???
        $answerids = array_values(array_map(create_function('$val',
         'return $val->id;'), $question->options->answers));
        // Shuffle the answers if required
        if (!empty($cmoptions->shuffleanswers) and !empty($question->options->shuffleanswers)) {
           $answerids = swapshuffle($answerids);
        }
        $state->options->order = $answerids;
        // Create empty responses
        $state->responses = array();
        return true;
    }

    function restore_session_and_responses(&$question, &$state) {
        // The serialized format for multiple choice quetsions
        // is an optional comma separated list of answer ids (the order of the
        // answers) followed by a colon, followed by another comma separated
        // list of answer ids, which are the radio/checkboxes that were
        // ticked.
        // E.g. 1,3,2,4:2,4 means that the answers were shown in the order
        // 1, 3, 2 and then 4 and the answers 2 and 4 were checked.

        $pos = strpos($state->responses[''], ':');
        if (false === $pos) { // No order of answers is given, so use the default
            $state->options->order = array_keys($question->options->answers);
        } else { // Restore the order of the answers
            $state->options->order = explode(',', substr($state->responses[''], 0, $pos));
            $state->responses[''] = substr($state->responses[''], $pos + 1);
        }
        // Restore the responses
        // This is done in different ways if only a single answer is allowed or
        // if multiple answers are allowed. For single answers the answer id is
        // saved in $state->responses[''], whereas for the multiple answers case
        // the $state->responses array is indexed by the answer ids and the
        // values are also the answer ids (i.e. key = value).
        if (empty($state->responses[''])) { // No previous responses
            $state->responses = array('' => '');
        } else {
            // Get array of answer ids
            $state->responses = explode(',', $state->responses['']);
            // Create an array indexed by these answer ids
            $state->responses = array_flip($state->responses);
            // Set the value of each element to be equal to the index
            array_walk($state->responses, create_function('&$a, $b',
             '$a = $b;'));
        }
        return true;
    }


    function get_correct_responses(&$question, &$state) {
        $responses = array();
        foreach ($question->options->answers as $answer) {
        if ($answer->fraction == 1) {
                $responses[$answer->id] = (string) $answer->id;
            }
        }
        return empty($responses) ? null : $responses;
    }

    function print_question_formulation_and_controls(&$question, &$state, $cmoptions, $options) {
        global $CFG;

        $answers = &$question->options->answers;
        $correctanswers = $this->get_correct_responses($question, $state);
        $readonly = empty($options->readonly) ? '' : 'disabled="disabled"';

        $formatoptions = new stdClass;
        $formatoptions->noclean = true;
        $formatoptions->para = false;

        // Print formulation
        $questiontext = format_text($question->questiontext,
                         $question->questiontextformat,
                         $formatoptions, $cmoptions->course);
        $image = get_question_image($question);
        $answerprompt = get_string('multipleanswers', 'quiz');

        // Print each answer in a separate row
        foreach ($state->options->order as $key => $aid) {
            $answer = &$answers[$aid];
            $checked = '';
            $chosen = false;

            $type = ' type="checkbox" ';
            $name   = "name=\"{$question->name_prefix}{$aid}\"";
            if (isset($state->responses[$aid])) {
                $checked = 'checked="checked"';
                $chosen = true;
            }

            $a = new stdClass;
            $a->id   = $question->name_prefix . $aid;
            $a->class = '';
            $a->feedbackimg = '';

            // Print the control
            $a->control = "<input $readonly id=\"$a->id\" $name $checked $type value=\"$aid\" />";

            if ($options->correct_responses) {
                $a->class = question_get_feedback_class(1);
            }
            if (($options->feedback && $chosen) || $options->correct_responses) {
                $a->feedbackimg = question_get_feedback_image($answer->fraction, $chosen && $options->feedback);
            }

            // Print the answer text
            $a->text = $this->number_in_style($key, $question->options->answernumbering) .
                    format_text($answer->answer, FORMAT_MOODLE, $formatoptions, $cmoptions->course);

            // Print feedback if feedback is on
            if (($options->feedback || $options->correct_responses) && $checked) {
                $a->feedback = format_text($answer->feedback, true, $formatoptions, $cmoptions->course);
            } else {
                $a->feedback = '';
            }

            $anss[] = clone($a);
        }

        $feedback = '';
        if ($options->feedback) {
            if ($state->raw_grade >= $question->maxgrade/1.01) {
                $feedback = $question->options->correctfeedback;
            } else {
                $feedback = $question->options->incorrectfeedback;
            }
            $feedback = format_text($feedback,
                    $question->questiontextformat,
                    $formatoptions, $cmoptions->course);
        }

        include("$CFG->dirroot/question/type/multichoiceset/display.html");
    }

    function grade_responses(&$question, &$state, $cmoptions) {
        $state->raw_grade = 0;
        $correctset = array();
        foreach ($question->options->answers as $answer) {
            if ($answer->fraction == 1) {
                array_push($correctset, $answer->id);
            }
        }

        $responseset = array();
        foreach ($state->responses as $response) {
            if ($response) {
                array_push($responseset, $response);
            }
        }

        // If the answer set and the response set match exactly, then the
        // response was correct
        if (count(array_diff($correctset, $responseset)) == 0 &&
            count(array_diff($responseset, $correctset)) == 0) {
            $state->raw_grade = 1;
        }

        // Make sure we don't assign negative or too high marks
        $state->raw_grade = min(max((float) $state->raw_grade,
                            0.0), 1.0) * $question->maxgrade;

        // Apply the penalty for this attempt
        $state->penalty = $question->penalty * $question->maxgrade;

        // mark the state as graded
        $state->event = ($state->event ==  QUESTION_EVENTCLOSE) ? QUESTION_EVENTCLOSEANDGRADE : QUESTION_EVENTGRADE;

        return true;
    }

/// BACKUP FUNCTIONS ////////////////////////////

    /*
     * Backup the data in the question
     *
     * This is used in question/backuplib.php
     */
    function backup($bf,$preferences,$question,$level=6) {

        $status = true;

        $multichoicesets = get_records("question_multichoiceset","question",$question,"id");
        //If there are multichoicesets
        if ($multichoicesets) {
            //Iterate over each multichoiceset
            foreach ($multichoicesets as $multichoiceset) {
                $status = fwrite ($bf,start_tag("MULTICHOICE",$level,true));
                //Print multichoiceset contents
                fwrite ($bf,full_tag("LAYOUT",$level+1,false,$multichoiceset->layout));
                fwrite ($bf,full_tag("ANSWERS",$level+1,false,$multichoiceset->answers));
                fwrite ($bf,full_tag("SHUFFLEANSWERS",$level+1,false,$multichoiceset->shuffleanswers));
                fwrite ($bf,full_tag("CORRECTFEEDBACK",$level+1,false,$multichoiceset->correctfeedback));
                fwrite ($bf,full_tag("INCORRECTFEEDBACK",$level+1,false,$multichoiceset->incorrectfeedback));
                fwrite ($bf,full_tag("ANSWERNUMBERING",$level+1,false,$multichoiceset->answernumbering));
                $status = fwrite ($bf,end_tag("MULTICHOICE",$level,true));
            }

            //Now print question_answers
            $status = question_backup_answers($bf,$preferences,$question);
        }
        return $status;
    }

/// RESTORE FUNCTIONS /////////////////

    /*
     * Restores the data in the question
     *
     * This is used in question/restorelib.php
     */
    function restore($old_question_id,$new_question_id,$info,$restore) {

        $status = true;

        //Get the multichoicesets array
        $multichoicesets = $info['#']['MULTICHOICE'];

        //Iterate over multichoicesets
        for($i = 0; $i < sizeof($multichoicesets); $i++) {
            $mul_info = $multichoicesets[$i];

            //Now, build the question_multichoiceset record structure
            $multichoiceset = new stdClass;
            $multichoiceset->question = $new_question_id;
            $multichoiceset->layout = backup_todb($mul_info['#']['LAYOUT']['0']['#']);
            $multichoiceset->answers = backup_todb($mul_info['#']['ANSWERS']['0']['#']);
            $multichoiceset->shuffleanswers = isset($mul_info['#']['SHUFFLEANSWERS']['0']['#'])?backup_todb($mul_info['#']['SHUFFLEANSWERS']['0']['#']):'';
            if (array_key_exists("CORRECTFEEDBACK", $mul_info['#'])) {
                $multichoiceset->correctfeedback = backup_todb($mul_info['#']['CORRECTFEEDBACK']['0']['#']);
            } else {
                $multichoiceset->correctfeedback = '';
            }
            if (array_key_exists("INCORRECTFEEDBACK", $mul_info['#'])) {
                $multichoiceset->incorrectfeedback = backup_todb($mul_info['#']['INCORRECTFEEDBACK']['0']['#']);
            } else {
                $multichoiceset->incorrectfeedback = '';
            }
            if (array_key_exists("ANSWERNUMBERING", $mul_info['#'])) {
                $multichoiceset->answernumbering = backup_todb($mul_info['#']['ANSWERNUMBERING']['0']['#']);
            } else {
                $multichoiceset->answernumbering = 'abc';
            }

            //We have to recode the answers field (a list of answers id)
            //Extracts answer id from sequence
            $answers_field = "";
            $in_first = true;
            $tok = strtok($multichoiceset->answers,",");
            while ($tok) {
                //Get the answer from backup_ids
                $answer = backup_getid($restore->backup_unique_code,"question_answers",$tok);
                if ($answer) {
                    if ($in_first) {
                        $answers_field .= $answer->new_id;
                        $in_first = false;
                    } else {
                        $answers_field .= ",".$answer->new_id;
                    }
                }
                //check for next
                $tok = strtok(",");
            }
            //We have the answers field recoded to its new ids
            $multichoiceset->answers = $answers_field;

            //The structure is equal to the db, so insert the question_shortanswer
            $newid = insert_record ("question_multichoiceset",$multichoiceset);

            //Do some output
            if (($i+1) % 50 == 0) {
                if (!defined('RESTORE_SILENTLY')) {
                    echo ".";
                    if (($i+1) % 1000 == 0) {
                        echo "<br />";
                    }
                }
                backup_flush(300);
            }

            if (!$newid) {
                $status = false;
            }
        }

        return $status;
    }

    /**
     * Decode links in question type specific tables.
     * @return bool success or failure.
     */
    function decode_content_links_caller($questionids, $restore, &$i) {
        $status = true;

        // Decode links in the question_multichoiceset table.
        if ($multichoicesets = get_records_list('question_multichoiceset', 'question',
                implode(',',  $questionids), '', 'id, correctfeedback, incorrectfeedback')) {

            foreach ($multichoicesets as $multichoiceset) {
                $correctfeedback = restore_decode_content_links_worker($multichoiceset->correctfeedback, $restore);
                $incorrectfeedback = restore_decode_content_links_worker($multichoiceset->incorrectfeedback, $restore);
                if ($correctfeedback != $multichoiceset->correctfeedback ||
                        $incorrectfeedback != $multichoiceset->incorrectfeedback) {
                    $subquestion->correctfeedback = addslashes($correctfeedback);
                    $subquestion->incorrectfeedback = addslashes($incorrectfeedback);
                    if (!update_record('question_multichoiceset', $multichoiceset)) {
                        $status = false;
                    }
                }

                // Do some output.
                if (++$i % 5 == 0 && !defined('RESTORE_SILENTLY')) {
                    echo ".";
                    if ($i % 100 == 0) {
                        echo "<br />";
                    }
                    backup_flush(300);
                }
            }
        }

        return $status;
    }

    function replace_file_links($question, $fromcourseid, $tocourseid, $url, $destination){
        parent::replace_file_links($question, $fromcourseid, $tocourseid, $url, $destination);
        // replace links in the question_match_sub table.
        // We need to use a separate object, because in load_question_options, $question->options->answers
        // is changed from a comma-separated list of ids to an array, so calling update_record on
        // $question->options stores 'Array' in that column, breaking the question.
        $optionschanged = false;
        $newoptions = new stdClass;
        $newoptions->id = $question->options->id;
        $newoptions->correctfeedback = question_replace_file_links_in_html($question->options->correctfeedback, $fromcourseid, $tocourseid, $url, $destination, $optionschanged);
        $newoptions->incorrectfeedback = question_replace_file_links_in_html($question->options->incorrectfeedback, $fromcourseid, $tocourseid, $url, $destination, $optionschanged);
        if ($optionschanged){
            if (!update_record('question_multichoiceset', addslashes_recursive($newoptions))) {
                error('Couldn\'t update \'question_multichoiceset\' record '.$newoptions->id);
            }
        }
        $answerchanged = false;
        foreach ($question->options->answers as $answer) {
            $answer->answer = question_replace_file_links_in_html($answer->answer, $fromcourseid, $tocourseid, $url, $destination, $answerchanged);
            if ($answerchanged){
                if (!update_record('question_answers', addslashes_recursive($answer))){
                    error('Couldn\'t update \'question_answers\' record '.$answer->id);
                }
            }
        }
    }

    /**
     * Runs all the code required to set up and save an essay question for testing purposes.
     * Alternate DB table prefix may be used to facilitate data deletion.
     */
    function generate_test($name, $courseid = null) {
        list($form, $question) = parent::generate_test($name, $courseid);
        $question->category = $form->category;
        $form->questiontext = "Which of these are planets in our solar system?";
        $form->generalfeedback = "General feedback";
        $form->penalty = 0.1;
        $form->shuffleanswers = 1;
        $form->answernumbering = 'abc';
        $form->noanswers = 3;
        $form->answer = array('Jupiter', 'Mars', 'Pluto');
        $form->fraction = array(1, 1, 0);
        $form->feedback = array('Good', 'Good', 'Pluto was reclassified as a dwarf planet in 2006.');
        $form->correctfeedback = 'Excellent!';
        $form->incorrectfeedback = 'Nope, try again!';

        if ($courseid) {
            $course = get_record('course', 'id', $courseid);
        }

        return $this->save_question($question, $form, $course);
    }

/// IMPORT EXPORT FUNCTIONS ////////////////////////////

    /**
     ** Provide export functionality for xml format
     ** @param question object the question object
     ** @param format object the format object so that helper methods can be used 
     ** @param extra mixed any additional format specific data that may be passed by the format (see format code for info)
     ** @return string the data to append to the output buffer or false if error
     **/
    function export_to_xml($question, $format, $extra) {
        $expout = '';

        $expout .= "    <correctfeedback>".$format->writetext($question->options->correctfeedback, 3)."</correctfeedback>\n";
        $expout .= "    <incorrectfeedback>".$format->writetext($question->options->incorrectfeedback, 3)."</incorrectfeedback>\n";
        $expout .= "    <answernumbering>{$question->options->answernumbering}</answernumbering>\n";
        foreach($question->options->answers as $answer) {
            $percent = $answer->fraction * 100;
            $expout .= "      <answer fraction=\"$percent\">\n";
            $expout .= $format->writetext($answer->answer,4,false);
            $expout .= "      <feedback>\n";
            $expout .= $format->writetext($answer->feedback,5,false);
            $expout .= "      </feedback>\n";
            $expout .= "    </answer>\n";
        }

        return $expout;
    }

   /**
    ** Provide import functionality for xml format
    ** @param data mixed the segment of data containing the question
    ** @param question object question object processed (so far) by standard imp
ort code
    ** @param format object the format object so that helper methods can be used
 (in particular error())
    ** @param extra mixed any additional format specific data that may be passed
 by the format (see format code for info)
    ** @return object question object suitable for save_options() call or false 
if cannot handle
    **/
   function import_from_xml($data, $question, $format, $extra=null) {
        // check question is for us
        $qtype = $data['@']['type'];
        if ($qtype=='multichoiceset') { 
            $qo = $format->import_headers($data);
            $qo->qtype = $qtype;

            $shuffleanswers = $format->getpath($data, array('#','shuffleanswers',0,'#'), 'false');
            $qo->answernumbering = $format->getpath($data, array('#','answernumbering',0,'#'), 'abc');
            $qo->shuffleanswers = $format->trans_single($shuffleanswers);
            $qo->correctfeedback = $format->getpath($data, array('#','correctfeedback',0,'#','text',0,'#'), '', true);
            $qo->incorrectfeedback = $format->getpath($data, array('#','incorrectfeedback',0,'#','text',0,'#'), '', true);
 
            // run through the answers
            $answers = $data['#']['answer'];
            $a_count = 0;
            foreach ($answers as $answer) {
                $ans = $format->import_answer($answer);
                $qo->answer[$a_count] = $ans->answer;
                $qo->fraction[$a_count] = $ans->fraction;
                $qo->feedback[$a_count] = $ans->feedback;
                ++$a_count;
            }
            return $qo;
        }
        else {
            return false;
        }
    } 

}
// Register this question type with the question bank.
question_register_questiontype(new question_multichoiceset_qtype());
?>
